@@ -1,17 +1,14 @@
 import os
-import sys
 import re
 import math
 import json
 import argparse
-import logging
 import base64
 import tqdm
 import skvideo.io
 import numpy as np
 import pandas as pd
 import cv2
-import radiomics
 import SimpleITK as sitk
 import PIL
 import PIL.Image
@@ -25,15 +22,21 @@ from multiprocessing import Manager, Pool
 from collections import OrderedDict
 from radiomics.featureextractor import RadiomicsFeatureExtractor
 from raymics import DatasetType
-from raymics.constants import COL_LABEL, RADIOMIC_FEATURE_FILENAME
+from raymics.constants import COL_LABEL, RADIOMIC_FEATURE_FILENAME, COL_INDEX,\
+    FILENAME_ANNOTATION, COL_DATA_PATH, COL_MASK_PATH, COL_ID
 from raymics.dataset import Dataset, LabeledDataset
 from raymics.utils import is_image, is_video, is_ndarray, is_labelme, \
     is_dicom_folder, isBase64
+from raymics.log import logger
+
+GROUP_LEN = 100
 
 
-logging.basicConfig(format="%(asctime)s - %(levelname)s: %(message)s",
-                    stream=sys.stdout, level=logging.ERROR)
-radiomics.logger.setLevel(logging.CRITICAL)
+def pop_unnamed(df: pd.DataFrame) -> pd.DataFrame:
+    unnamed_cols = [col for col in df.columns if col.startswith("Unnamed")]
+    for col in unnamed_cols:
+        df.pop(col)
+    return df
 
 
 def dummy_mask(shape: Tuple[int, int]) -> sitk.Image:
@@ -207,26 +210,30 @@ def read_mask(mask_path: str) -> sitk.Image:
     return mask
 
 
-def execute(extractor, data, mask, label=None) -> OrderedDict:
-    feature_data = extractor.execute(data, mask)
-    if label is not None:
-        feature_data[COL_LABEL] = label
-    return feature_data
+def extract_non_ts(data_idxs: List[int],
+                   data_ids: List,
+                   data_paths: List[str],
+                   mask_paths: List[str],
+                   result_labels_path: str,
+                   extractor: RadiomicsFeatureExtractor, lock):
 
+    def read_non_ts(data_path, mask_path):
 
-def extract_non_ts(data_paths: List[str], mask_paths: List[str],
-                   labels: List[Union[int, float, int, None]],
-                   result_path: str, extractor: RadiomicsFeatureExtractor, lock):
-
-    def read_non_ts(data_path, mask_path, label):
         # opencv image
         if is_image(data_path):
-            # img = cv_img2sitk_image(data_path)
-            img = sitk.ReadImage(data_path, sitk.sitkInt8)
+            try:
+                img = sitk.ReadImage(data_path, sitk.sitkInt8)
+            except Exception as e:
+                logger.info(f"Exception: {e}")
+                logger.info(f"Failed to read image using sitk: {data_path}.")
+                logger.info(f"Try to read the image using opencv.")
+                img = cv_img2sitk_image(data_path)
+
         # numpy ndarray
         elif is_ndarray(data_path):
             arr = np.load(data_path)
             img = sitk.GetImageFromArray(arr)
+
         # dicom folder
         elif is_dicom_folder(data_path):
             with tempfile.TemporaryDirectory() as tmpdir:
@@ -234,29 +241,60 @@ def extract_non_ts(data_paths: List[str], mask_paths: List[str],
                     os.path.join(tmpdir, f"{string.ascii_lowercase}.nii.gz")
                 dicom2nifti.dicom_series_to_nifti(data_path, tmp_img_path)
                 img = sitk.ReadImage(tmp_img_path)
+
         # try to read image using SimpleITK
         else:
             img = sitk.ReadImage(data_path)
+
         mask = get_mask(mask_path=mask_path,
                         shape=(img.GetHeight(), img.GetWidth()))
 
-        return img, mask, label
+        # to avoiding ITK ERROR:
+        # LabelStatisticsImageFilter(0x1263e0bb0): inputs do not occupy
+        # the same physical space!
+        mask.SetOrigin(img.GetOrigin())
+        mask.SetSpacing(img.GetSpacing())
+        mask.SetDirection(img.GetDirection())
+
+        return img, mask
 
     feature_list: List[OrderedDict] = [
-        execute(extractor, *read_non_ts(data, mask, label))
-        for data, mask, label in zip(data_paths, mask_paths, labels)
+        extractor.execute(*read_non_ts(data, mask))
+        for data, mask in zip(data_paths, mask_paths)
     ]
+
     feature_df = pd.DataFrame(feature_list)
+    columns = feature_df.columns.tolist()
+    if data_ids:
+        feature_df[COL_ID] = data_ids
+        feature_df = feature_df[[COL_ID] + columns]
+    else:
+        feature_df = feature_df[columns]
+
+    feature_df[COL_INDEX] = data_idxs
+    feature_df.set_index(COL_INDEX)
 
     with lock:
-        if not os.path.exists(result_path):
-            feature_df.to_csv(result_path, index=False)
+        if not os.path.exists(result_labels_path):
+            feature_df.to_csv(result_labels_path,
+                              index=True,
+                              index_label=COL_INDEX)
         else:
-            feature_df.to_csv(result_path, index=False, mode="a", header=False)
+            feature_df.to_csv(result_labels_path,
+                              index=True,
+                              mode="a",
+                              header=False)
 
 
-def extract_ts(data_path: str, mask_path: str, label: Union[int, float, int, None],
-               result_path: str, extractor: RadiomicsFeatureExtractor):
+def extract_ts(data_idx: int,
+               data_path: str,
+               mask_path: str,
+               result_dir: str,
+               result_path: str,
+               extractor: RadiomicsFeatureExtractor,
+               original_labels_path: str,
+               result_labels_path: str,
+               lock):
 
     def read_ts():
         mask = None
@@ -267,29 +305,50 @@ def extract_ts(data_path: str, mask_path: str, label: Union[int, float, int, Non
                 mask = get_mask(mask_path=mask_path,
                                 shape=(img.GetHeight(), img.GetWidth()))
 
-            yield img, mask, label
+            yield img, mask
     try:
-        feature_list: List[OrderedDict] = [execute(extractor, data, mask, label)
-                                           for data, mask, label in read_ts()]
+        feature_list: List[OrderedDict] = \
+            [extractor.execute(data, mask) for data, mask in read_ts()]
         feature_df = pd.DataFrame(feature_list)
 
         if os.path.exists(result_path):
             result_path = rename_duplicate_path(result_path)
 
         feature_df.to_csv(result_path, index=False)
+
+        with lock:
+            if not os.path.exists(result_labels_path):
+                result_labels_df = pd.read_csv(original_labels_path)
+                for col in [COL_ID, COL_DATA_PATH, COL_MASK_PATH]:
+                    if col in result_labels_df:
+                        result_labels_df.pop(col)
+                result_labels_df[COL_DATA_PATH] = None
+                result_labels_df[COL_INDEX] = list(range(result_labels_df.shape[0]))
+                result_labels_df.set_index(COL_INDEX)
+            else:
+                result_labels_df = pd.read_csv(result_labels_path,
+                                               index_col=COL_INDEX)
+
+            rel_path = os.path.relpath(result_path, result_dir)
+            result_labels_df[COL_DATA_PATH].loc[data_idx] = rel_path
+
+            pop_unnamed(result_labels_df).to_csv(result_labels_path,
+                                                 index=True,
+                                                 index_label=COL_INDEX)
+
     except Exception as e:
-        logging.warning(f"\nRadiomics extracting error: {e}, "
-                        f"\ndata_path: {data_path}, "
-                        f"\nmask_path: {mask_path}")
+        logger.warning(f"Radiomics extracting error: {e}, "
+                       f"\ndata_path: {data_path}, "
+                       f"\nmask_path: {mask_path}")
 
 
-def extract(data_dir: str, result_dir: str, config: Union[str, dict],
+def extract(dataset_dir: str, result_dir: str, config: Union[str, dict],
             processes: int = 2, progress = None) -> RadiomicsFeatureExtractor:
     """
 
     Parameters
     ----------
-    data_dir : str
+    dataset_dir : str
         Directory of the raw dataset.
 
     result_dir : str
@@ -302,30 +361,30 @@ def extract(data_dir: str, result_dir: str, config: Union[str, dict],
         Processes to extracting radiomics features.
 
     """
-    dataset = LabeledDataset(root_dir=data_dir)
+    dataset = LabeledDataset(root_dir=dataset_dir)
 
     if not os.path.exists(result_dir):
         os.makedirs(result_dir)
 
-    if not os.path.exists(result_dir):
-        os.makedirs(result_dir)
+    result_labels_path = os.path.join(result_dir, FILENAME_ANNOTATION)
 
     pool = Pool(processes=processes)
     extractor = RadiomicsFeatureExtractor(config)
+    # force to output without additional info
+    extractor.settings["additionalInfo"] = False
 
-    data_type = DatasetType.TS if is_video(dataset[0][0]) else DatasetType.NON_TS
+    data_type = DatasetType.TS if is_video(dataset[0][2]) else DatasetType.NON_TS
+
+    lock = Manager().Lock()
 
     # non-time series
     if data_type == DatasetType.NON_TS:
-        result_path = os.path.join(result_dir, RADIOMIC_FEATURE_FILENAME)
-        assert not os.path.exists(result_path), \
-            f"The result_path is not supposed to exist: {result_path}"
 
-        group_len = 1000
         idxs = list(range(len(dataset)))
-        group_idxs = [idxs[i: i + group_len] for i in range(0, len(idxs), group_len)]
-
-        lock = Manager().Lock()
+        group_idxs = [idxs[i: i + GROUP_LEN]
+                      for i in range(0, len(idxs), GROUP_LEN)]
+        group_data_idxs = [dataset.data_idxs[i: i + GROUP_LEN]
+                           for i in range(0, len(idxs), GROUP_LEN)]
 
         bar = tqdm.tqdm(total=len(group_idxs))
         if progress is None:
@@ -334,17 +393,71 @@ def extract(data_dir: str, result_dir: str, config: Union[str, dict],
             progress.set_tqdm(bar=bar)
             callback = progress.update
 
-        for idx_list in group_idxs:
-            pool.apply_async(
-                func=extract_non_ts,
-                args=([dataset.data_paths[_] for _ in idx_list],
-                      [dataset.mask_paths[_] for _ in idx_list],
-                      [dataset.labels[_] for _ in idx_list],
-                      result_path,
-                      extractor,
-                      lock),
-                callback=callback
-            )
+        for data_idxs, idx_list in zip(group_data_idxs, group_idxs):
+            extract_non_ts(data_idxs,
+                           [dataset.data_ids[_] for _ in idx_list]
+                           if dataset.data_ids else None,
+                           [dataset.data_paths[_] for _ in idx_list],
+                           [dataset.mask_paths[_] for _ in idx_list],
+                           result_labels_path,
+                           extractor,
+                           lock)
+        # TODO: run with multi-process
+        #
+        #     pool.apply_async(
+        #         func=extract_non_ts,
+        #         args=(group_data_idxs,
+        #               [dataset.data_ids[_] for _ in idx_list]
+        #               if dataset.data_ids else None,
+        #               [dataset.data_paths[_] for _ in idx_list],
+        #               [dataset.mask_paths[_] for _ in idx_list],
+        #               result_labels_path,
+        #               extractor,
+        #               lock),
+        #         callback=callback
+        #     )
+        #
+        # pool.close()
+        # pool.join()
+
+        # concat original labels.csv and feature labels.csv
+        labels_df = pd.read_csv(dataset.labels_file)
+        labels_df[COL_INDEX] = list(range(labels_df.shape[0]))
+        labels_df.set_index(COL_INDEX)
+        labels_cols = labels_df.columns.tolist()
+
+        feature_df = pd.read_csv(result_labels_path)
+        feature_df.set_index(COL_INDEX)
+        any_feature_col = [col for col in feature_df.columns
+                           if col not in [COL_ID, COL_DATA_PATH, COL_MASK_PATH,
+                                          COL_INDEX, COL_LABEL]][0]
+
+        df = pd.concat([labels_df, feature_df], axis=1)
+        for col in [COL_ID, COL_MASK_PATH] + \
+                   [col for col in df.columns if col.startswith(COL_INDEX)]:
+            if col in df:
+                df.pop(col)
+
+        # labels.csv
+        dff = df[~df[any_feature_col].isna()]
+        labels_cols = [col for col in dff.columns if col in labels_cols]
+        if COL_LABEL in dff:
+            labels_cols = [COL_DATA_PATH, COL_LABEL] + \
+                          [col for col in labels_cols if col != COL_DATA_PATH
+                           and col != COL_LABEL]
+        else:
+            labels_cols = [COL_DATA_PATH] + \
+                          [col for col in labels_cols if col != COL_DATA_PATH]
+        dff[labels_cols].to_csv(result_labels_path, index=False)
+
+        # radiomics data
+        result_features_path = os.path.join(result_dir,
+                                            RADIOMIC_FEATURE_FILENAME)
+        dff.pop(COL_DATA_PATH)
+        if COL_LABEL in dff:
+            dff.pop(COL_LABEL)
+        feature_cols = [col for col in dff.columns if col not in labels_cols]
+        dff[feature_cols].to_csv(result_features_path, index=False)
 
     # time series
     else:    # data_type == 1
@@ -360,17 +473,30 @@ def extract(data_dir: str, result_dir: str, config: Union[str, dict],
             progress.set_tqdm(bar=bar)
             callback = progress.update
 
-        for i, (data_path, mask_path, label) in enumerate(dataset):
+        for i in range(len(dataset)):
+            data_idx, data_id, data_path, mask_path, label = dataset[i]
             result_path = result_paths[i]
-            # extract_ts(data_path, mask_path, label, result_path, extractor)
+            # extract_ts(data_idx, data_path, mask_path, result_dir,
+            #            result_path, extractor, dataset.labels_file,
+            #            result_labels_path, lock)
             pool.apply_async(
                 func=extract_ts,
-                args=(data_path, mask_path, label, result_path, extractor),
+                args=(data_idx, data_path, mask_path, result_dir,
+                      result_path, extractor, dataset.labels_file,
+                      result_labels_path, lock),
                 callback=callback
             )
 
-    pool.close()
-    pool.join()
+        pool.close()
+        pool.join()
+
+        # remove index col and nan data_paths
+        df = pd.read_csv(result_labels_path, index_col=COL_INDEX)
+        df = df[~df[COL_DATA_PATH].isna()]
+        columns = [COL_DATA_PATH] + \
+                  [_ for _ in df.columns if _ != COL_DATA_PATH and
+                   not _.startswith(COL_INDEX)]
+        df[columns].to_csv(result_labels_path, index=False)
 
     return extractor
 
@@ -393,14 +519,10 @@ if __name__ == "__main__":
         type=str,
         help="yaml or json file path, or base64 string of json file."
     )
-    parser.add_argument(
-        "--cloud_run",
-        action='store_true',
-        help="Whether run in cloud."
-    )
     parser.add_argument(
         "--processes",
         type=int,
+        default=2,
         help="multiprocessing pool max size."
     )
 
@@ -413,9 +535,9 @@ if __name__ == "__main__":
     else:
         assert os.path.exists(config), f"config file not exists: {config}"
 
-    logging.info(f"config: {config}")
+    logger.info(f"config: {config}")
 
-    extract(data_dir=args.data_dir,
+    extract(dataset_dir=args.dataset_dir,
             config=config,
             result_dir=args.result_dir,
             processes=args.processes)
